#define EOVERFLOW E2BIG
#define ENOSYS EINVAL

#define HAVE_VSNPRINTF 1
#define HAVE_ERRNO_H 1
#define HAVE_HTONS 1
#define HAVE_ARPA_INET_H 1

/* Enable ELF visibility */
#define CACA_ENABLE_VISIBILITY 1
