#ifndef __GRAPHICS_H
#define __GRAPHICS_H

// These are the rasterizer opcodes.
enum {
    ROP_JMP,
    ROP_NOP,
    ROP_SETLINECOLOR,
    ROP_SETLINEPATTERN,
    ROP_SETLINEWIDTH,
    ROP_SETFILLCOLOR,
    ROP_SETFILLPATTERN,
    ROP_UNDEFINED,
    ROP_SETTEXTSIZE,
    ROP_SETTEXTALIGN,
    ROP_SETTEXTANGLE,
    ROP_SETTEXTSTYLE,
    ROP_SETMARKERSIZE,
    ROP_SETMARKERSYMBOL,
    ROP_CLEARRECTANGLE,
    ROP_DRAWSINGLELINE,
    ROP_DRAWRECTANGLEEDGES,
    ROP_DRAWRECTANGLEINTERIOR,
    ROP_DRAWRECTANGLE,
    ROP_DRAWSLICEEDGES,
    ROP_DRAWSLICEINTERIOR,
    ROP_DRAWSLICE,
    ROP_DRAWARC,
    ROP_DRAWPOLYLINE,
    ROP_DRAWPOLYGONINTERIOR,
    ROP_DRAWPOLYGON,
    ROP_DRAWTEXT,
    ROP_DRAWMARKER,
};

#endif
