.text
.type __require_ref,@function
__require_ref:
    call    a64l@plt
    call    abort@plt
    call    abs@plt
    call    access@plt
    call    acct@plt
    call    alarm@plt
    call    asctime@plt
    call    atof@plt
    call    atoi@plt
    call    atol@plt
    call    bsearch@plt
    call    calloc@plt
    call    chdir@plt
    call    chmod@plt
    call    chroot@plt
    call    clearerr@plt
    call    clock@plt
    call    closedir@plt
    call    copysign@plt
    call    creat@plt
    call    ctermid@plt
    call    ctime@plt
    call    cuserid@plt
    call    daylight@plt
    call    drand48@plt
    call    dup2@plt
    call    dup@plt
    call    eaccess@plt
    call    ecvt@plt
    call    endgrent@plt
    call    endpwent@plt
    call    endutent@plt
    call    erand48@plt
    call    execle@plt
    call    execl@plt
    call    execlp@plt
    call    execve@plt
    call    execv@plt
    call    execvp@plt
    call    _exit@plt
    call    exit@plt
    call    fcvt@plt
    call    free@plt
    call    frexp@plt
    call    fstatfs@plt
    call    fstat@plt
    call    ftok@plt
    call    ftw@plt
    call    gcvt@plt
    call    getchar@plt
    call    getcwd@plt
    call    getegid@plt
    call    getenv@plt
    call    geteuid@plt
    call    getuid@plt
    call    getgid@plt
    call    getgrent@plt
    call    getgrgid@plt
    call    getgrnam@plt
    call    getlogin@plt
    call    getopt@plt
    call    getpass@plt
    call    getpgrp@plt
    call    getpid@plt
    call    getppid@plt
    call    getpwent@plt
    call    getpwnam@plt
    call    getpwuid@plt
    call    getutent@plt
    call    getutid@plt
    call    getutline@plt
    call    getw@plt
    call    gmtime@plt
    call    gsignal@plt
    call    hcreate@plt
    call    hdestroy@plt
    call    hsearch@plt
    call    isatty@plt
    call    jrand48@plt
    call    kill@plt
    call    l64a@plt
    call    labs@plt
    call    lcong48@plt
    call    ldexp@plt
    call    lfind@plt
    call    link@plt
    call    localtime@plt
    call    lockf@plt
    call    lrand48@plt
    call    lsearch@plt
    call    lseek@plt
    call    mallinfo@plt
    call    malloc@plt
    call    mallopt@plt
    call    _mcount@plt
    call    memccpy@plt
    call    memchr@plt
    call    memcmp@plt
    call    memcpy@plt
    call    memmove@plt
    call    memset@plt
    call    mkdir@plt
    call    mknod@plt
    call    mlock@plt
    call    modf@plt
    call    mount@plt
    call    mrand48@plt
    call    msgget@plt
    call    msgrcv@plt
    call    msgsnd@plt
    call    nice@plt
    call    nrand48@plt
    call    opendir@plt
    call    open@plt
    call    pause@plt
    call    perror@plt
    call    poll@plt
    call    printf@plt
    call    profil@plt
    call    ptrace@plt
    call    putchar@plt
    call    putenv@plt
    call    putpwent@plt
    call    puts@plt
    call    pututline@plt
    call    putw@plt
    call    qsort@plt
    call    rand@plt
    call    readdir@plt
    call    read@plt
    call    realloc@plt
    call    rename@plt
    call    rewind@plt
    call    rmdir@plt
    call    sbrk@plt
    call    scanf@plt
    call    seed48@plt
    call    seekdir@plt
    call    semget@plt
    call    semop@plt
    call    setbuf@plt
    call    setgid@plt
    call    setgrent@plt
    call    setpgrp@plt
    call    setpwent@plt
    call    setuid@plt
    call    setutent@plt
    call    setvbuf@plt
    call    shmat@plt
    call    shmdt@plt
    call    shmget@plt
    call    sighold@plt
    call    sigignore@plt
    call    signal@plt
    call    sigpause@plt
    call    sigrelse@plt
    call    sigset@plt
    call    sleep@plt
    call    sprintf@plt
    call    srand48@plt
    call    srand@plt
    call    sscanf@plt
    call    ssignal@plt
    call    statfs@plt
    call    stat@plt
    call    strcat@plt
    call    strchr@plt
    call    strcmp@plt
    call    strcpy@plt
    call    strcspn@plt
    call    strdup@plt
    call    strlen@plt
    call    strncat@plt
    call    strncmp@plt
    call    strncpy@plt
    call    strpbrk@plt
    call    strrchr@plt
    call    strspn@plt
    call    strtod@plt
    call    strtok@plt
    call    strtol@plt
    call    swab@plt
    call    sync@plt
    call    syscall@plt
    call    system@plt
    call    tdelete@plt
    call    telldir@plt
    call    tfind@plt
    call    time@plt
    call    times@plt
    call    timezone@plt
    call    tolower@plt
    call    toupper@plt
    call    tsearch@plt
    call    ttyname@plt
    call    ttyslot@plt
    call    twalk@plt
    call    tzname@plt
    call    tzset@plt
    call    ulimit@plt
    call    umask@plt
    call    umount@plt
    call    uname@plt
    call    ungetc@plt
    call    unlink@plt
    call    utime@plt
    call    utmpname@plt
    call    vfprintf@plt
    call    vprintf@plt
    call    vsprintf@plt
    call    wait@plt
    call    write@plt
    call    close@plt
    call    tgetent@plt
    call    tgetstr@plt
    call    tgetflag@plt
    call    tgetnum@plt
    call    tgoto@plt
    call    tputs@plt
    call    tmpfile@plt
    call    chown@plt
    call    msgctl@plt
    call    semctl@plt
    call    shmctl@plt
    call    fstat64@plt
    call    stat64@plt
    call    atexit@plt
    call    sin@plt
    call    cos@plt
    call    tan@plt
    call    atan2@plt
    call    asin@plt
    call    acos@plt
    call    remainder@plt
    call    sqrt@plt
    call    log@plt
    call    log10@plt
    call    floor@plt
    call    ceil@plt
    call    fmod@plt
    call    fabs@plt
