#ifndef __LOTERRS_H
#define __LOTERRS_H

enum {
    LOTERR_FILENAME_TOO_LONG    = 9083,
    LOTERR_FILENAME_INVALID     = 9248,
    LOTERR_WORKSHEET_UNNAMED    = 9130,
    LOTERR_FILE_IN_USE          = 9043,
    LOTERR_CELL_IMMUTABLE       = 9034,
    LOTERR_MEMORY_FULL          = 8962,
};

#endif
